import * as wasm from './lalrpop_lambda_bg';

const heap = new Array(32);

heap.fill(undefined);

heap.push(undefined, null, true, false);

let heap_next = heap.length;

function addHeapObject(obj) {
    if (heap_next === heap.length) heap.push(heap.length + 1);
    const idx = heap_next;
    heap_next = heap[idx];

    heap[idx] = obj;
    return idx;
}

let cachedTextDecoder = new TextDecoder('utf-8');

let cachegetUint8Memory = null;
function getUint8Memory() {
    if (cachegetUint8Memory === null || cachegetUint8Memory.buffer !== wasm.memory.buffer) {
        cachegetUint8Memory = new Uint8Array(wasm.memory.buffer);
    }
    return cachegetUint8Memory;
}

function getStringFromWasm(ptr, len) {
    return cachedTextDecoder.decode(getUint8Memory().subarray(ptr, ptr + len));
}

let cachedGlobalArgumentPtr = null;
function globalArgumentPtr() {
    if (cachedGlobalArgumentPtr === null) {
        cachedGlobalArgumentPtr = wasm.__wbindgen_global_argument_ptr();
    }
    return cachedGlobalArgumentPtr;
}

let cachegetUint32Memory = null;
function getUint32Memory() {
    if (cachegetUint32Memory === null || cachegetUint32Memory.buffer !== wasm.memory.buffer) {
        cachegetUint32Memory = new Uint32Array(wasm.memory.buffer);
    }
    return cachegetUint32Memory;
}

export function __wbindgen_string_new(p, l) { return addHeapObject(getStringFromWasm(p, l)); }

function getObject(idx) { return heap[idx]; }

export function __wbindgen_number_get(n, invalid) {
    let obj = getObject(n);
    if (typeof(obj) === 'number') return obj;
    getUint8Memory()[invalid] = 1;
    return 0;
}

export function __wbindgen_boolean_get(i) {
    let v = getObject(i);
    return typeof(v) === 'boolean' ? (v ? 1 : 0) : 2;
}

let WASM_VECTOR_LEN = 0;

let cachedTextEncoder = new TextEncoder('utf-8');

let passStringToWasm;
if (typeof cachedTextEncoder.encodeInto === 'function') {
    passStringToWasm = function(arg) {


        let size = arg.length;
        let ptr = wasm.__wbindgen_malloc(size);
        let offset = 0;
        {
            const mem = getUint8Memory();
            for (; offset < arg.length; offset++) {
                const code = arg.charCodeAt(offset);
                if (code > 0x7F) break;
                mem[ptr + offset] = code;
            }
        }

        if (offset !== arg.length) {
            arg = arg.slice(offset);
            ptr = wasm.__wbindgen_realloc(ptr, size, size = offset + arg.length * 3);
            const view = getUint8Memory().subarray(ptr + offset, ptr + size);
            const ret = cachedTextEncoder.encodeInto(arg, view);

            offset += ret.written;
        }
        WASM_VECTOR_LEN = offset;
        return ptr;
    };
} else {
    passStringToWasm = function(arg) {


        let size = arg.length;
        let ptr = wasm.__wbindgen_malloc(size);
        let offset = 0;
        {
            const mem = getUint8Memory();
            for (; offset < arg.length; offset++) {
                const code = arg.charCodeAt(offset);
                if (code > 0x7F) break;
                mem[ptr + offset] = code;
            }
        }

        if (offset !== arg.length) {
            const buf = cachedTextEncoder.encode(arg.slice(offset));
            ptr = wasm.__wbindgen_realloc(ptr, size, size = offset + buf.length);
            getUint8Memory().set(buf, ptr + offset);
            offset += buf.length;
        }
        WASM_VECTOR_LEN = offset;
        return ptr;
    };
}

export function __wbindgen_string_get(i, len_ptr) {
    let obj = getObject(i);
    if (typeof(obj) !== 'string') return 0;
    const ptr = passStringToWasm(obj);
    getUint32Memory()[len_ptr / 4] = WASM_VECTOR_LEN;
    return ptr;
}

function dropObject(idx) {
    if (idx < 36) return;
    heap[idx] = heap_next;
    heap_next = idx;
}

function takeObject(idx) {
    const ret = getObject(idx);
    dropObject(idx);
    return ret;
}

export function __wbindgen_rethrow(idx) { throw takeObject(idx); }

export function __wbindgen_throw(ptr, len) {
    throw new Error(getStringFromWasm(ptr, len));
}

function freeExp(ptr) {

    wasm.__wbg_exp_free(ptr);
}
/**
* A parsed λ-expression
*
* This struct is a wrapper around an [`Expression`] to both allow exporting it to JS, and to
* contain functions functions we want to export for the `Exp`, for example [`Exp::to_string`].
*/
export class Exp {

    static __wrap(ptr) {
        const obj = Object.create(Exp.prototype);
        obj.ptr = ptr;

        return obj;
    }

    free() {
        const ptr = this.ptr;
        this.ptr = 0;
        freeExp(ptr);
    }

    /**
    * Parse and construct a new `Expr`
    *
    * ```js
    * new lambda.Exp(\"(\\\\x.x x) y\");
    * new lambda.Exp(2);
    * new lambda.Exp(false);
    * new lambda.Exp(\"*wtf\");  // Throws exception.
    * ```
    * @param {any} v
    * @returns {}
    */
    constructor(v) {
        this.ptr = wasm.exp_new(addHeapObject(v));
    }
    /**
    * @param {boolean} η
    * @returns {Exp}
    */
    applicative(η) {
        return Exp.__wrap(wasm.exp_applicative(this.ptr, η));
    }
    /**
    * @returns {Exp}
    */
    call_by_value() {
        return Exp.__wrap(wasm.exp_call_by_value(this.ptr));
    }
    /**
    * @param {boolean} η
    * @returns {Exp}
    */
    normal(η) {
        return Exp.__wrap(wasm.exp_normal(this.ptr, η));
    }
    /**
    * @returns {Exp}
    */
    call_by_name() {
        return Exp.__wrap(wasm.exp_call_by_name(this.ptr));
    }
    /**
    * @param {boolean} η
    * @returns {Exp}
    */
    head_spine(η) {
        return Exp.__wrap(wasm.exp_head_spine(this.ptr, η));
    }
    /**
    * See [`std::fmt::Display`]
    *
    * ```js
    * let expr = new lambda.Exp(\"\\\\x.x x\");
    * console.log(`${expr}`);
    * ```
    * @returns {string}
    */
    toString() {
        const retptr = globalArgumentPtr();
        wasm.exp_toString(retptr, this.ptr);
        const mem = getUint32Memory();
        const rustptr = mem[retptr / 4];
        const rustlen = mem[retptr / 4 + 1];

        const realRet = getStringFromWasm(rustptr, rustlen).slice();
        wasm.__wbindgen_free(rustptr, rustlen * 1);
        return realRet;

    }
    /**
    * See `From<Expression> for u64`
    *
    * ```js
    * let two = new lambda.Exp(\"\\\\f.\\\\x.(f (f x))\");
    * console.log(`${two.toNumber()}`);
    * ```
    * @returns {number}
    */
    toNumber() {
        return wasm.exp_toNumber(this.ptr) >>> 0;
    }
    /**
    * See `From<Expression> for bool`
    *
    *
    * ```js
    * let t = new lambda.Exp(\"\\\\a.\\\\b.a\");
    * console.log(`${t.toBool()}`);
    * ```
    * @returns {boolean}
    */
    toBool() {
        return (wasm.exp_toBool(this.ptr)) !== 0;
    }
}

export function __wbindgen_object_drop_ref(i) { dropObject(i); }

